use super::data::{vec_get, vec_size};
use crate::errors;
use libc::{c_char, size_t};
use log;
use std;
use std::error::Error as StdError;
use std::ffi::CString;

pub struct Error {
    pub msg: CString,
    pub kind: CString,
}

pub type ErrorList = Vec<Error>;

struct CauseIterator<'a> {
    current: Option<&'a StdError>,
}

impl<'a> std::iter::Iterator for CauseIterator<'a> {
    type Item = Error;

    fn next(&mut self) -> std::option::Option<Error> {
        let std_error = self.current?;
        let result = Error {
            msg: CString::new(std_error.to_string()).unwrap_or(CString::default()),
            kind: CString::new("Cause").unwrap_or(CString::default()),
        };
        self.current = std_error.source();
        Some(result)
    }
}

fn error_kind(e: &errors::Error) -> &str {
    match e {
        errors::Error::AQLSyntaxError { .. } => "AQLSyntaxError",
        errors::Error::AQLSemanticError { .. } => "AQLSemanticError",
        errors::Error::LoadingGraphFailed { .. } => "LoadingGraphFailed",
        errors::Error::ImpossibleSearch(_) => "ImpossibleSearch",
        errors::Error::NoSuchCorpus(_) => "NoSuchCorpus",
        errors::Error::Generic { .. } => "Generic",
        errors::Error::IO(_) => "IO",
        errors::Error::Bincode(_) => "Bincode",
        errors::Error::CSV(_) => "CSV",
        errors::Error::ParseIntError(_) => "ParseIntError",
        errors::Error::Fmt(_) => "Fmt",
        errors::Error::Strum(_) => "Strum",
        errors::Error::Regex(_) => "Regex",
    }
}

impl From<errors::Error> for ErrorList {
    fn from(e: errors::Error) -> ErrorList {
        let mut result = ErrorList::new();
        result.push(Error {
            msg: CString::new(e.to_string()).unwrap_or(CString::default()),
            kind: CString::new(error_kind(&e)).unwrap_or(CString::default()),
        });
        let cause_it = CauseIterator {
            current: e.source(),
        };
        for e in cause_it {
            result.push(e)
        }
        return result;
    }
}

impl From<log::SetLoggerError> for Error {
    fn from(e: log::SetLoggerError) -> Error {
        let err = if let Ok(error_msg) = CString::new(e.to_string()) {
            Error {
                msg: error_msg,
                kind: CString::new("SetLoggerError").unwrap(),
            }
        } else {
            // meta-error
            Error {
                msg: CString::new(String::from("Some error occurred")).unwrap(),
                kind: CString::new("SetLoggerError").unwrap(),
            }
        };
        return err;
    }
}

impl From<std::io::Error> for Error {
    fn from(e: std::io::Error) -> Error {
        let err = if let Ok(error_msg) = CString::new(e.to_string()) {
            Error {
                msg: error_msg,
                kind: CString::new("std::io::Error").unwrap(),
            }
        } else {
            // meta-error
            Error {
                msg: CString::new(String::from("Some error occurred")).unwrap(),
                kind: CString::new("std::io::Error").unwrap(),
            }
        };
        return err;
    }
}

pub fn new(err: errors::Error) -> *mut ErrorList {
    Box::into_raw(Box::new(ErrorList::from(err)))
}

#[no_mangle]
pub extern "C" fn annis_error_size(ptr: *const ErrorList) -> size_t {
    vec_size(ptr)
}

#[no_mangle]
pub extern "C" fn annis_error_get_msg(ptr: *const ErrorList, i: size_t) -> *const c_char {
    let item = vec_get(ptr, i);
    if item.is_null() {
        return std::ptr::null();
    }
    let err: &Error = cast_const!(item);
    return err.msg.as_ptr();
}

#[no_mangle]
pub extern "C" fn annis_error_get_kind(ptr: *const ErrorList, i: size_t) -> *const c_char {
    let item = vec_get(ptr, i);
    if item.is_null() {
        return std::ptr::null();
    }
    let err: &Error = cast_const!(item);
    return err.kind.as_ptr();
}
